use super::{
    compiler::Compiler,
    expr::{Expr, ExprOp, Variable},
    opcode::OpCode,
    parser::{parse_query, Query},
    runtime::execute,
    value::{Value, ValueOrRef},
    Addr, Error, Register,
};
use crate::Transaction;
use std::collections::HashMap;

#[derive(Debug, Clone)]
pub struct Filter {
    /// Query
    query: Query,
    /// List of external variables
    variables: Vec<Variable>,
    /// Opcodes with human readable labels
    opcodes: Vec<OpCode>,
    /// The list of opcodes that make up the program, the labels has been converted into addresses
    opcodes_to_execute: Vec<OpCode>,
    /// If the program has some boilerplate that can be skipped. This is non-zero when the initial
    /// program has been executed at compile time, and the register has been populated to
    /// `initial_register`. Everytime a new Runtime is created, instead of executing the program,
    /// it can safely start-off a different Address, and clone the initial_register.
    start_at: Addr,
    /// The state of the register
    initial_register: Vec<Value>,
}

impl Filter {
    pub fn new(code: &str) -> Result<Self, Error> {
        let query = parse_query(code)?;
        let opcodes = query.where_clause.as_ref().map_or_else(
            || {
                Ok(vec![
                    OpCode::LOAD(0.into(), true.into()),
                    OpCode::HLT(0.into()),
                ])
            },
            |expr| Compiler::new(expr).compile(),
        )?;

        Ok(Self {
            query,
            opcodes: opcodes.clone(),
            variables: opcodes
                .iter()
                .filter_map(|x| match x {
                    OpCode::LOAD_EXTERNAL(_, name) => Some(name.clone()),
                    _ => None,
                })
                .collect(),
            opcodes_to_execute: Compiler::resolve_label_to_addr(opcodes)?,
            start_at: 0.into(),
            initial_register: vec![],
        })
    }

    /// Attempts to optiomize the `raw_opcodes` inside the Filter. Returns a tuple with the new
    /// self with the optimized opcodes, and a boolean indicating if the program was optimized.
    pub fn optimize_round(mut self) -> (Self, bool) {
        let (new_opcodes, has_changed) = Compiler::optimize(self.opcodes);
        self.opcodes = new_opcodes;
        self.opcodes_to_execute = Compiler::resolve_label_to_addr(self.opcodes.clone()).unwrap();

        (self, has_changed)
    }

    /// Returns a human readable version of the compiled program (generated op-codes)
    pub fn debug(&self) -> String {
        self.opcodes
            .iter()
            .map(|x| match x {
                OpCode::HLT(_) | OpCode::LABEL(_) => x.to_string(),
                x => format!("\t{}", x.to_string()),
            })
            .collect::<Vec<_>>()
            .join("\n")
    }

    pub fn dump(&self) -> String {
        self.opcodes_to_execute
            .iter()
            .enumerate()
            .map(|(pos, opcode)| format!("{}: {}", pos, opcode.to_string()))
            .collect::<Vec<_>>()
            .join("\n")
    }

    pub fn execute<'a>(
        &'a self,
        external_variables: &'a HashMap<Variable, ValueOrRef<'a>>,
    ) -> Result<Value, Error> {
        execute(
            external_variables,
            &self.opcodes_to_execute,
            self.initial_register.clone(),
            self.start_at,
        )
    }
}

#[cfg(test)]
mod test {
    use super::Filter;
    use crate::filter_expr::{
        opcode::OpCode,
        value::{Value, ValueOrRef},
    };

    #[test]
    fn parse() {
        let mut x = Filter::new(
            r#"
            WHERE
                $foo = 3 + 2 * 4 / 2 * 298210 + $bar
                AND 25 = 5*$five
        "#,
        )
        .unwrap();
        println!("{}\n", x.dump());

        loop {
            let (new_x, is_done) = x.optimize_round();
            println!("{}\n", new_x.dump());
            x = new_x;
            if !is_done {
                break;
            }
        }

        let external_variables_1 = vec![
            ("foo".into(), ValueOrRef::Value(0.into())),
            ("bar".into(), ValueOrRef::Value(0.into())),
            ("five".into(), ValueOrRef::Value(5.into())),
        ]
        .into_iter()
        .collect();

        let external_variables_2 = vec![
            ("foo".into(), ValueOrRef::Value(1192844.into())),
            ("bar".into(), ValueOrRef::Value(1.into())),
            ("five".into(), ValueOrRef::Value(4.into())),
        ]
        .into_iter()
        .collect();

        let external_variables_3 = vec![
            ("foo".into(), ValueOrRef::Value(1192844.into())),
            ("bar".into(), ValueOrRef::Value(1.into())),
            ("five".into(), ValueOrRef::Value(5.into())),
        ]
        .into_iter()
        .collect();

        let value = x.execute(&external_variables_1).expect("valid execution");
        assert!(matches!(value, Value::Bool(false)));

        let value = x.execute(&external_variables_2).expect("valid execution");
        assert!(matches!(value, Value::Bool(false)));

        let value = x.execute(&external_variables_3).expect("valid execution");
        assert!(matches!(value, Value::Bool(true)));
    }
}
